@@ -39,11 +39,10 @@ from starlette.types import ASGIApp
 from starlette.websockets import WebSocket, WebSocketDisconnect
 from yarl import URL
 
-from satori.const import Api
+from satori.const import Api, EventType
 from satori.model import Event, Meta, ModelBase, Opcode
 from satori.utils import decode
 
-from .. import EventType
 from .adapter import Adapter as Adapter
 from .connection import WebsocketConnection
 from .formdata import parse_content_disposition as parse_content_disposition
@@ -60,6 +59,16 @@ _T_ws_endpoint = TypeVar("_T_ws_endpoint", bound=Callable[[WebSocket], Awaitable
 StarletteResponse = Response
 
 
+async def _json(self: StarletteRequest) -> Any:
+    if not hasattr(self, "_json"):
+        body = await self.body()
+        self._json = decode(body)
+    return self._json
+
+
+StarletteRequest.json = _json
+
+
 async def _request_handler(action: str, request: StarletteRequest, func: RouteCall, platform: str, self_id: str):
     if action == Api.UPLOAD_CREATE.value:
         async with request.form() as form:
@@ -74,11 +83,15 @@ async def _request_handler(action: str, request: StarletteRequest, func: RouteCa
             )
             return JSONResponse(content=res)
     try:
+        if request.method == "GET":
+            params = dict(request.query_params)
+        else:
+            params = await request.json()
         res = await func(
             Request(
                 request,
                 action,
-                decode(await request.body()),
+                params,
                 platform=platform,
                 self_id=self_id,
             )
@@ -298,17 +311,27 @@ class Server(Service, RouterMixin):
         self_id: str = request.headers.get("X-Self-ID") or request.headers.get("Satori-User-ID")  # type: ignore
 
         for _router in self._adapters:
-            if action not in _router.routes:
+            if action in _router.routes:
+                func = _router.routes[action]
+            elif action.startswith("internal") and "internal/*" in _router.routes:
+                func = _router.routes["internal/*"]
+            else:
                 continue
             if not _router.ensure(platform, self_id):
                 continue
-            return await _request_handler(action, request, _router.routes[action], platform, self_id)
+            return await _request_handler(action, request, func, platform, self_id)
         if action in self.routes:
             return await _request_handler(action, request, self.routes[action], platform, self_id)
+        if action.startswith("internal") and "internal/*" in self.routes:
+            return await _request_handler(action, request, self.routes["internal/*"], platform, self_id)
         for _router in self.routers:
-            if action not in _router.routes:
+            if action in _router.routes:
+                func = _router.routes[action]
+            elif action.startswith("internal") and "internal/*" in _router.routes:
+                func = _router.routes["internal/*"]
+            else:
                 continue
-            return await _request_handler(action, request, _router.routes[action], platform, self_id)
+            return await _request_handler(action, request, func, platform, self_id)
         return Response(status_code=404, content=action)
 
     async def proxy_url_handler(self, request: StarletteRequest):
@@ -320,7 +343,7 @@ class Server(Service, RouterMixin):
                 isinstance(resp, (PlainTextResponse, HTMLResponse, JSONResponse)) or resp.__class__ is Response
             ) and len(resp.body) > self.stream_threshold:
 
-                async def iter_content(body: bytes):
+                async def iter_content(body: bytes | memoryview[int]):
                     for i in range(0, len(body), self.stream_chunk_size):
                         yield body[i : i + self.stream_chunk_size]
 
@@ -407,7 +430,7 @@ class Server(Service, RouterMixin):
         return JSONResponse(content=Meta(logins, proxy_urls).dump())
 
     async def webhook_create_handler(self, request: StarletteRequest):
-        body = decode(await request.body())
+        body = await request.json()
         url = body["url"]
         token = body.get("token")
         self.webhooks.append(WebhookEndpoint(url, token))
@@ -427,7 +450,7 @@ class Server(Service, RouterMixin):
         return Response()
 
     async def webhook_delete_handler(self, request: StarletteRequest):
-        body = decode(await request.body())
+        body = await request.json()
         url = body["url"]
         for endpoint in self.webhooks:
             if endpoint.url == url:
